# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2011-2016 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for `filestore.protocols` module.
"""

from unittest import TestCase
import os
from random import SystemRandom

from skein import skein512
from dbase32 import db32enc, db32dec

from filestore import protocols, misc


random = SystemRandom()
MiB = 1024 * 1024
TwoMiB = 2 * MiB
EightMiB = 8 * MiB
GiB = 1024 * MiB
COUNT = 50 * 1000
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'


class DigestAccum(set):
    def add(self, digest):
        if not isinstance(digest, bytes):
            raise TypeError('digest must be bytes')
        if not (25 <= len(digest) <= 35 and len(digest) % 5 == 0):
            raise ValueError('digest has bad length {}'.format(len(digest)))
        if digest in self:
            raise ValueError('digest already in set')
        super().add(digest)
        return digest


class TestAccum(TestCase):
    def test_add(self):
        accum = DigestAccum()
        self.assertEqual(accum, set())
        self.assertEqual(len(accum), 0)

        with self.assertRaises(TypeError) as cm:
            accum.add('A' * 30)
        self.assertEqual(str(cm.exception), 'digest must be bytes')

        for size in [24, 26, 29, 31, 34, 36]:
            bad = os.urandom(size)
            with self.assertRaises(ValueError) as cm:
                accum.add(bad)
            self.assertEqual(
                str(cm.exception),
                'digest has bad length {}'.format(size)
            )

        d1 = os.urandom(30)
        d2 = os.urandom(30)

        self.assertIs(accum.add(d1), d1)
        self.assertEqual(accum, set([d1]))
        self.assertEqual(len(accum), 1)

        self.assertIs(accum.add(d2), d2)
        self.assertEqual(accum, set([d1, d2]))
        self.assertEqual(len(accum), 2)

        with self.assertRaises(ValueError) as cm:
            accum.add(d1)
        self.assertEqual(str(cm.exception), 'digest already in set')

        self.assertEqual(accum, set([d1, d2]))
        self.assertEqual(len(accum), 2)


class TestConstants(TestCase):
    def test_PERS(self):
        all_pers = (
            protocols.PERS_LEAF,
            protocols.PERS_ROOT,
        )
        for pers in all_pers:
            self.assertIsInstance(pers, bytes)
            self.assertEqual(pers[:36],
                b'20110430 jderose@novacut.com dmedia/'
            )
        self.assertEqual(len(set(all_pers)), len(all_pers))

        # Be extra safe the what's actually used in V1:
        self.assertIsInstance(protocols.PERS_LEAF, bytes)
        self.assertIsInstance(protocols.PERS_ROOT, bytes)
        self.assertNotEqual(protocols.PERS_LEAF, protocols.PERS_ROOT)

    def test_MiB(self):
        self.assertIsInstance(protocols.MiB, int)
        self.assertEqual(protocols.MiB, 1024 * 1024)
        self.assertEqual(protocols.MiB, 2**20)

    def test_MAX_FILE_SIZE(self):
        # For now, constrain to JavaScript integer limit (2**53):
        self.assertIsInstance(protocols.MAX_FILE_SIZE, int)
        self.assertEqual(protocols.MAX_FILE_SIZE, 2**53)

    def test_MIN_LEAF_SIZE(self):
        self.assertIsInstance(protocols.MIN_LEAF_SIZE, int)
        self.assertEqual(protocols.MIN_LEAF_SIZE, 2 * 1024 * 1024)
        self.assertEqual(protocols.MIN_LEAF_SIZE, 2**21)
        self.assertEqual(protocols.MIN_LEAF_SIZE,
            protocols.MAX_FILE_SIZE // 2**32
        )


class TestFunctions(TestCase):
    def test_make_key(self):
        with self.assertRaises(TypeError) as cm:
            protocols.make_key(1.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('integer', int, float, 1.0)
        )
        with self.assertRaises(TypeError) as cm:
            protocols.make_key('1')
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('integer', int, str, '1')
        )

        with self.assertRaises(ValueError) as cm:
            protocols.make_key(-1)
        self.assertEqual(
            str(cm.exception),
            'integer: must be >= 0; got -1'
        )
        with self.assertRaises(ValueError) as cm:
            protocols.make_key(-17)
        self.assertEqual(
            str(cm.exception),
            'integer: must be >= 0; got -17'
        )

        self.assertEqual(protocols.make_key(0), b'0')
        self.assertEqual(protocols.make_key(1), b'1')
        self.assertEqual(protocols.make_key(1776), b'1776')
        for i in range(1000):
            n = random.randint(0, 16 * GiB)
            self.assertEqual(protocols.make_key(n), repr(n).encode('ascii'))


class BaseTestCase1(TestCase):
    def setUp(self):
        self._checks = []

    def check_hash_leaf_validation(self, proto):
        # Test with wrong leaf_index type:
        with self.assertRaises(TypeError) as cm:
            proto.hash_leaf(0.0, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 0.0)
        )
        with self.assertRaises(TypeError) as cm:
            proto.hash_leaf(1.5, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 1.5)
        )

        # Test when leaf_index is too small:
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(-1, None)
        self.assertEqual(
            str(cm.exception),
            'Need 0 <= leaf_index < {}; got -1'.format(proto.max_leaf_count)
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(-17, None)
        self.assertEqual(
            str(cm.exception),
            'Need 0 <= leaf_index < {}; got -17'.format(proto.max_leaf_count)
        )

        # Test when leaf_index is too big:
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(proto.max_leaf_count, None)
        self.assertEqual(
            str(cm.exception),
            'Need 0 <= leaf_index < {}; got {}'.format(
                proto.max_leaf_count, proto.max_leaf_count
            )
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(proto.max_leaf_count + 1, None)
        self.assertEqual(
            str(cm.exception),
            'Need 0 <= leaf_index < {}; got {}'.format(
                proto.max_leaf_count, proto.max_leaf_count + 1
            )
        )

        # Test with wrong leaf_data type
        leaf_data = 'a' * 31415
        with self.assertRaises(TypeError) as cm:
            proto.hash_leaf(2, leaf_data)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_data', bytes, str, leaf_data)
        )

        # Test with wrong leaf_data length
        toosmall = b''
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(2, toosmall)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= {}; got 0'.format(proto.leaf_size)
        )
        toobig = b'a' * (proto.leaf_size + 1)
        with self.assertRaises(ValueError) as cm:
            proto.hash_leaf(2, toobig)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= {}; got {}'.format(
                proto.leaf_size, proto.leaf_size + 1
            )
        )

        self._checks.append('hash_leaf_validation')

    def check_hash_root_validation(self, proto):
        # Test with wrong file_size type:
        with self.assertRaises(TypeError) as cm:
            proto.hash_root(1.0, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.0)
        )
        with self.assertRaises(TypeError) as cm:
            proto.hash_root(1.5, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.5)
        )

        # Test when file_size is too small:
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(0, None)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= file_size <= 9007199254740992; got 0'
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(-1, None)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= file_size <= 9007199254740992; got -1'
        )

        # Test when file_size is too big:
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(2**53 + 1, None)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= file_size <= 9007199254740992; got 9007199254740993'
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(2**64 - 1, None)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= file_size <= 9007199254740992; got 18446744073709551615'
        )

        # Test with wrong leaf_hashes type:
        leaf_hashes = 'A' * proto.digest_bytes
        with self.assertRaises(TypeError) as cm:
            proto.hash_root(17, leaf_hashes)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_hashes', bytes, str, leaf_hashes)
        )

        # Test with empty leaf_hashes
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(17, b'')
        self.assertEqual(str(cm.exception), 'leaf_hashes cannot be empty')

        # Test when len(leaf_hashes) is not a multiple of Protocol.digest_bytes
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(17, b'A' * (proto.digest_bytes + 1))
        self.assertEqual(
            str(cm.exception),
            'len(leaf_hashes) is {}, not multiple of {}'.format(
                proto.digest_bytes + 1, proto.digest_bytes
            )
        )

        # Test low and high bounds
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(proto.leaf_size + 1, b'a' * proto.digest_bytes)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= file_size <= {}; got {}'.format(
                proto.leaf_size, proto.leaf_size + 1
            )
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(proto.leaf_size, b'ab' * proto.digest_bytes)
        self.assertEqual(
            str(cm.exception),
            'Need {} <= file_size <= {}; got {}'.format(
                proto.leaf_size + 1,
                2 * proto.leaf_size,
                proto.leaf_size
            )
        )
        with self.assertRaises(ValueError) as cm:
            proto.hash_root(
                2 * proto.leaf_size + 1,
                b'ab' * proto.digest_bytes
            )
        self.assertEqual(
            str(cm.exception),
            'Need {} <= file_size <= {}; got {}'.format(
                proto.leaf_size + 1,
                2 * proto.leaf_size,
                2 * proto.leaf_size + 1
            )
        )

        self._checks.append('hash_root_validation')


class TestProtocol(BaseTestCase1):
    def test_init(self):
        # Wrong `leaf_size` type:
        with self.assertRaises(TypeError) as cm:
            protocols.Protocol(1.0 * TwoMiB, 200)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_size', int, float, 1.0 * TwoMiB)
        )

        # leaf_size < MIN_LEAF_SIZE:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB - 1, 200)
        self.assertEqual(
            str(cm.exception),
            'leaf_size: must be >= 2097152; got 2097151'
        )

        # leaf_size % MIN_LEAF_SIZE != 0:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB * 3 // 2, 200)
        self.assertEqual(
            str(cm.exception),
            'leaf_size: must be multiple of 2097152; got 3145728'
        )

        # Wrong `digest_bits` type:
        with self.assertRaises(TypeError) as cm:
            protocols.Protocol(TwoMiB, 200.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('digest_bits', int, float, 200.0)
        )

        # digest_bits < 200:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB, 159)
        self.assertEqual(
            str(cm.exception),
            'digest_bits: must be >= 200; got 159'
        )

        # digest_bits % 40 != 0:
        with self.assertRaises(ValueError) as cm:
            protocols.Protocol(TwoMiB, 220)
        self.assertEqual(
            str(cm.exception),
            'digest_bits: must be multiple of 40; got 220'
        )

        # Test with the good minimum values:
        proto = protocols.Protocol(TwoMiB, 200)
        self.assertEqual(proto.leaf_size, TwoMiB)
        self.assertEqual(proto.max_leaf_count,
             protocols.MAX_FILE_SIZE // TwoMiB
        )
        self.assertEqual(proto.max_leaf_count, 2**32)
        self.assertEqual(proto.digest_bits, 200)
        self.assertEqual(proto.digest_bytes, 25)
        self.assertEqual(proto.digest_b32len, 40)

        # Test with the actual v1 protocol values:
        proto = protocols.Protocol(8 * MiB, 240)
        self.assertEqual(proto.leaf_size, 8 * MiB)
        self.assertEqual(proto.max_leaf_count, 1073741824)
        self.assertEqual(proto.max_leaf_count, 2**30)
        self.assertEqual(proto.digest_bits, 240)
        self.assertEqual(proto.digest_bytes, 30)
        self.assertEqual(proto.digest_b32len, 48)

        # Test with possible future values:
        proto = protocols.Protocol(16 * MiB, 280)
        self.assertEqual(proto.leaf_size, 16 * MiB)
        self.assertEqual(proto.max_leaf_count, 536870912)
        self.assertEqual(proto.max_leaf_count, 2**29)
        self.assertEqual(proto.digest_bits, 280)
        self.assertEqual(proto.digest_bytes, 35)
        self.assertEqual(proto.digest_b32len, 56)

    def test_repr(self):
        proto = protocols.Protocol(TwoMiB, 200)
        self.assertEqual(repr(proto),
            'Protocol(2097152, 200)'
        )
        proto = protocols.Protocol(8 * MiB, 240)
        self.assertEqual(repr(proto),
            'Protocol(8388608, 240)'
        )

    def test_hash_leaf(self):
        proto = protocols.Protocol(TwoMiB, 200)
        self.check_hash_leaf_validation(proto)
        with self.assertRaises(NotImplementedError) as cm:
            proto.hash_leaf(0, b'data')
        self.assertEqual(
            str(cm.exception),
            'Protocol._hash_leaf(key_leaf_index, leaf_data, challenge)'
        )
        with self.assertRaises(NotImplementedError) as cm:
            proto.hash_leaf(0, b'data', b'secret')
        self.assertEqual(
            str(cm.exception),
            'Protocol._hash_leaf(key_leaf_index, leaf_data, challenge)'
        )
        with self.assertRaises(NotImplementedError) as cm:
            proto._hash_leaf(0, b'data', b'secret')
        self.assertEqual(
            str(cm.exception),
            'Protocol._hash_leaf(key_leaf_index, leaf_data, challenge)'
        )

        class Example(protocols.Protocol):
            pass

        proto = Example(TwoMiB, 200)
        self.check_hash_leaf_validation(proto)
        with self.assertRaises(NotImplementedError) as cm:
            proto.hash_leaf(0, b'data')
        self.assertEqual(
            str(cm.exception),
            'Example._hash_leaf(key_leaf_index, leaf_data, challenge)'
        )
        with self.assertRaises(NotImplementedError) as cm:
            proto.hash_leaf(0, b'data', b'secret')
        self.assertEqual(
            str(cm.exception),
            'Example._hash_leaf(key_leaf_index, leaf_data, challenge)'
        )
        with self.assertRaises(NotImplementedError) as cm:
            proto._hash_leaf(0, b'data', b'secret')
        self.assertEqual(
            str(cm.exception),
            'Example._hash_leaf(key_leaf_index, leaf_data, challenge)'
        )

        class FooProtocol(protocols.Protocol):
            def _hash_leaf(self, *args):
                assert not hasattr(self, '_call')
                assert not hasattr(self, '_ret')
                self._call = args
                self._ret = os.urandom(self.digest_bytes)
                return self._ret

        foo = FooProtocol(TwoMiB, 200)
        self.check_hash_leaf_validation(foo)
        digest = foo.hash_leaf(0, b'data')
        self.assertEqual(foo._call, (b'0', b'data', b''))
        self.assertIs(digest, foo._ret)

        foo = FooProtocol(TwoMiB, 200)
        digest = foo.hash_leaf(17, b'more data', b'secret')
        self.assertEqual(foo._call, (b'17', b'more data', b'secret'))
        self.assertIs(digest, foo._ret)

        for i in range(100):
            index = random.randint(0, MiB)
            length = random.randint(1, TwoMiB)
            data = os.urandom(1) * length

            foo = FooProtocol(TwoMiB, 200)
            digest = foo.hash_leaf(index, data)
            self.assertEqual(foo._call,
                (str(index).encode(), data, b'')
            )
            self.assertIs(digest, foo._ret)

            nonce = os.urandom(16)
            foo = FooProtocol(TwoMiB, 200)
            digest = foo.hash_leaf(index, data, nonce)
            self.assertEqual(foo._call,
                (str(index).encode(), data, nonce)
            )
            self.assertIs(digest, foo._ret)

    def test_hash_root(self):
        proto = protocols.Protocol(TwoMiB, 200)
        self.check_hash_root_validation(proto)
        with self.assertRaises(NotImplementedError) as cm:
            proto.hash_root(1, b'A' * 25)
        self.assertEqual(
            str(cm.exception),
            'Protocol._hash_root(key_file_size, leaf_hashes)'
        )
        with self.assertRaises(NotImplementedError) as cm:
            proto._hash_root(1, b'A' * 25)
        self.assertEqual(
            str(cm.exception),
            'Protocol._hash_root(key_file_size, leaf_hashes)'
        )

        class Example(protocols.Protocol):
            pass

        proto = Example(TwoMiB, 200)
        self.check_hash_root_validation(proto)
        with self.assertRaises(NotImplementedError) as cm:
            proto.hash_root(1, b'A' * 25)
        self.assertEqual(
            str(cm.exception),
            'Example._hash_root(key_file_size, leaf_hashes)'
        )
        with self.assertRaises(NotImplementedError) as cm:
            proto._hash_root(1, b'A' * 25)
        self.assertEqual(
            str(cm.exception),
            'Example._hash_root(key_file_size, leaf_hashes)'
        )

        class FooProtocol(protocols.Protocol):
            def _hash_root(self, *args):
                assert not hasattr(self, '_call')
                assert not hasattr(self, '_ret')
                self._call = args
                self._ret = os.urandom(self.digest_bytes)
                return self._ret

        foo = FooProtocol(TwoMiB, 200)
        self.check_hash_root_validation(foo)
        digest = foo.hash_root(1, b'A' * 25)
        self.assertEqual(foo._call, (b'1', b'A' * 25))
        self.assertIs(digest, foo._ret)

        for i in range(100):
            size = random.randint(1, GiB)
            count = size // TwoMiB
            if size % TwoMiB:
                count += 1
            data = os.urandom(25) * count

            foo = FooProtocol(TwoMiB, 200)
            digest = foo.hash_root(size, data)
            self.assertEqual(foo._call,
                (str(size).encode(), data)
            )
            self.assertIs(digest, foo._ret)


class BaseTestCase2(BaseTestCase1):
    def check_hash_leaf_simple(self, proto):
        accum = DigestAccum()
        small = b'D'  # Smallest possible leaf
        large = b'D' * proto.leaf_size  # Largest possible leaf

        # Smallest leaf, index=0
        digest = accum.add(proto.hash_leaf(0, small))
        self.assertEqual(proto._hash_leaf(b'0', small, b''), digest)

        # Smallest leaf, index=1
        digest = accum.add(proto.hash_leaf(1, small))
        self.assertEqual(proto._hash_leaf(b'1', small, b''), digest)

        # Largest leaf, index=0
        digest = accum.add(proto.hash_leaf(0, large))
        self.assertEqual(proto._hash_leaf(b'0', large, b''), digest)

        # Largest leaf, index=1
        digest = accum.add(proto.hash_leaf(1, large))
        self.assertEqual(proto._hash_leaf(b'1', large, b''), digest)

        # Test with challenge, smallest leaf, index=0:
        digest = accum.add(proto.hash_leaf(0, small, b'secret'))
        self.assertEqual(proto._hash_leaf(b'0', small, b'secret'), digest)

        # Test with challenge, smallest leaf, index=1:
        digest = accum.add(proto.hash_leaf(1, small, b'secret'))
        self.assertEqual(proto._hash_leaf(b'1', small, b'secret'), digest)

        # Test with challenge, largest leaf, index=0:
        digest = accum.add(proto.hash_leaf(0, large, b'secret'))
        self.assertEqual(proto._hash_leaf(b'0', large, b'secret'), digest)

        # Test with challenge, largest leaf, index=1:
        digest = accum.add(proto.hash_leaf(1, large, b'secret'))
        self.assertEqual(proto._hash_leaf(b'1', large, b'secret'), digest)

        # Make sure we didn't goof
        self.assertEqual(len(accum), 8)
        for digest in accum:
            self.assertEqual(len(digest), proto.digest_bytes)
        safety = set()
        for index in [0, 1]:
            for data in [small, large]:
                for challenge in [b'', b'secret']:
                    safety.add(proto.hash_leaf(index, data, challenge))
        self.assertEqual(safety, accum)

        self._checks.append('hash_leaf_simple')

    def check_hash_root_simple(self, proto):
        accum = DigestAccum()
        one = b'D' * proto.digest_bytes
        two = b'AB' * proto.digest_bytes

        digest = accum.add(proto.hash_root(1, one))
        self.assertEqual(proto._hash_root(b'1', one), digest)

        digest = accum.add(proto.hash_root(2, one))
        self.assertEqual(proto._hash_root(b'2', one), digest)

        size = proto.leaf_size + 1
        digest = accum.add(proto.hash_root(size, two))
        self.assertEqual(
            proto._hash_root(str(size).encode(), two),
            digest
        )

        size = proto.leaf_size + 2
        digest = accum.add(proto.hash_root(size, two))
        self.assertEqual(
            proto._hash_root(str(size).encode(), two),
            digest
        )

        size = proto.leaf_size * 2
        digest = accum.add(proto.hash_root(size, two))
        self.assertEqual(
            proto._hash_root(str(size).encode(), two),
            digest
        )

        # Make sure we didn't goof
        self.assertEqual(len(accum), 5)
        for digest in accum:
            self.assertEqual(len(digest), proto.digest_bytes)
        safety = set([
            proto.hash_root(1, one),
            proto.hash_root(2, one),
            proto.hash_root(proto.leaf_size + 1, two),
            proto.hash_root(proto.leaf_size + 2, two),
            proto.hash_root(proto.leaf_size * 2, two),
        ])
        self.assertEqual(safety, accum)

        self._checks.append('hash_root_simple')

    def check_hash_leaf_crypto(self, proto):
        """
        Random value sanity check on Protocol.hash_leaf().
        """

        # Sanity check on our cryptographic claim that the
        # leaf-hash is tied to the leaf-index:
        leaf_data = os.urandom(16)
        accum = set(
            proto.hash_leaf(i, leaf_data, b'')
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)

        # Sanity check on our cryptographic claim that the
        # leaf-hash is tied to the leaf-data:
        accum = set(
            proto.hash_leaf(21, os.urandom(16), b'')
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)

        # Sanity check on our cryptographic claim that the
        # leaf-hash is tied to the challenge:
        accum = set(
            proto.hash_leaf(21, leaf_data, os.urandom(16))
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)

        self._checks.append('hash_leaf_crypto')

    def check_hash_root_crypto(self, proto):
        """
        Random value sanity check on Protocol.hash_root().
        """

        # Sanity check on our cryptographic claim that the
        # root-hash is tied to the file-size:
        leaf_hashes = os.urandom(proto.digest_bytes)
        accum = set(
            proto.hash_root(size, leaf_hashes)
            for size in range(1, COUNT + 1)
        )
        self.assertEqual(len(accum), COUNT)

        # Sanity check on our cryptographic claim that the
        # root-hash is tied to the leaf-hashes:
        accum = set(
            proto.hash_root(314159, os.urandom(proto.digest_bytes))
            for i in range(COUNT)
        )
        self.assertEqual(len(accum), COUNT)

        self._checks.append('hash_root_crypto')

    def check_hash_leaf(self, proto):
        self.check_hash_leaf_validation(proto)
        self.check_hash_leaf_simple(proto)
        self.check_hash_leaf_crypto(proto)

    def check_hash_root(self, proto):
        self.check_hash_root_validation(proto)
        self.check_hash_root_simple(proto)
        self.check_hash_root_crypto(proto)


class TestSkeinProtocol(BaseTestCase2):
    def test_repr(self):
        proto = protocols.SkeinProtocol(TwoMiB, 200)
        self.assertEqual(repr(proto),
            'SkeinProtocol(2097152, 200)'
        )

    def test_hash_leaf(self):
        proto = protocols.SkeinProtocol(TwoMiB, 200)
        self.check_hash_leaf(proto)
        self.assertEqual(self._checks, [
            'hash_leaf_validation',
            'hash_leaf_simple',
            'hash_leaf_crypto',
        ])

        accum = DigestAccum()
        small = b'A'  # Smallest possible leaf
        large = b'B' * proto.leaf_size  # Largest possible leaf

        # Smallest leaf, index=0
        digest = accum.add(proto.hash_leaf(0, small))
        self.assertEqual(digest,
            skein512(small,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'0', small, b''), digest)

        # Smallest leaf, index=1
        digest = accum.add(proto.hash_leaf(1, small))
        self.assertEqual(digest,
            skein512(small,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'1', small, b''), digest)

        # Largest leaf, index=0
        digest = accum.add(proto.hash_leaf(0, large))
        self.assertEqual(digest,
            skein512(large,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'0', large, b''), digest)

        # Largest leaf, index=1
        digest = accum.add(proto.hash_leaf(1, large))
        self.assertEqual(digest,
            skein512(large,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'1', large, b''), digest)

        # Test with challenge, smallest leaf, index=0:
        digest = accum.add(proto.hash_leaf(0, small, b'secret'))
        self.assertEqual(digest,
            skein512(small,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
                nonce=b'secret',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'0', small, b'secret'), digest)

        # Test with challenge, smallest leaf, index=1:
        digest = accum.add(proto.hash_leaf(1, small, b'secret'))
        self.assertEqual(digest,
            skein512(small,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
                nonce=b'secret',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'1', small, b'secret'), digest)

        # Test with challenge, largest leaf, index=0:
        digest = accum.add(proto.hash_leaf(0, large, b'secret'))
        self.assertEqual(digest,
            skein512(large,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'0',
                nonce=b'secret',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'0', large, b'secret'), digest)

        # Test with challenge, largest leaf, index=1:
        digest = accum.add(proto.hash_leaf(1, large, b'secret'))
        self.assertEqual(digest,
            skein512(large,
                digest_bits=200,
                pers=protocols.PERS_LEAF,
                key=b'1',
                nonce=b'secret',
            ).digest()
        )
        self.assertEqual(proto._hash_leaf(b'1', large, b'secret'), digest)

        # Make sure we didn't goof
        safety = set()
        for index in [0, 1]:
            for data in [small, large]:
                for challenge in [b'', b'secret']:
                    safety.add(proto.hash_leaf(index, data, challenge))
        self.assertEqual(safety, accum)
        self.assertEqual(len(accum), 8)
        for digest in accum:
            self.assertEqual(len(digest), 25)

    def test_hash_root(self):
        proto = protocols.SkeinProtocol(TwoMiB, 200)
        self.check_hash_root(proto)
        self.assertEqual(self._checks, [
            'hash_root_validation',
            'hash_root_simple',
            'hash_root_crypto',
        ])

        accum = DigestAccum()
        one = b'D' * 25
        two = b'D' * 50

        # Smallest file-size possible for one leaf:
        digest = accum.add(proto.hash_root(1, one))
        self.assertEqual(digest,
            skein512(one,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'1',
            ).digest()
        )
        self.assertEqual(proto._hash_root(b'1', one), digest)

        # Largest file-size possible for one leaf:
        digest = accum.add(proto.hash_root(TwoMiB, one))
        self.assertEqual(digest,
            skein512(one,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'2097152',
            ).digest()
        )
        self.assertEqual(proto._hash_root(b'2097152', one), digest)

        # Smallest file-size possible for two leaves:
        digest = accum.add(proto.hash_root(TwoMiB + 1, two))
        self.assertEqual(digest,
            skein512(two,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'2097153',
            ).digest()
        )
        self.assertEqual(proto._hash_root(b'2097153', two), digest)

        # Largest file-size possible for two leaves:
        digest = accum.add(proto.hash_root(2 * TwoMiB, two))
        self.assertEqual(digest,
            skein512(two,
                digest_bits=200,
                pers=protocols.PERS_ROOT,
                key=b'4194304',
            ).digest()
        )
        self.assertEqual(proto._hash_root(b'4194304', two), digest)

        # Make sure we didn't goof
        self.assertEqual(len(accum), 4)
        safety = set([
            proto.hash_root(1, one),
            proto.hash_root(TwoMiB, one),
            proto.hash_root(TwoMiB + 1, two),
            proto.hash_root(TwoMiB * 2, two),
        ])
        self.assertEqual(safety, accum)
        for digest in accum:
            self.assertEqual(len(digest), 25)


class Test_VERSION1(BaseTestCase2):
    proto = protocols.VERSION1

    def test_repr(self):
        self.assertEqual(
            repr(protocols.VERSION1),
            'SkeinProtocol(8388608, 240)'
        )

    def test_hash_leaf(self):
        self.check_hash_leaf(self.proto)
        self.assertEqual(self._checks, [
            'hash_leaf_validation',
            'hash_leaf_simple',
            'hash_leaf_crypto',
        ])

        leaves = misc.build_leaves(EightMiB)
        vectors = misc.load_data('V1')
        self.assertEqual(
            db32enc(self.proto.hash_leaf(0, leaves['A'])),
            vectors['leaf_hashes']['A'][0]
        )
        self.assertEqual(
            db32enc(self.proto.hash_leaf(1, leaves['A'])),
            vectors['leaf_hashes']['A'][1]
        )
        self.assertEqual(
            db32enc(self.proto.hash_leaf(0, leaves['B'])),
            vectors['leaf_hashes']['B'][0]
        )
        self.assertEqual(
            db32enc(self.proto.hash_leaf(1, leaves['B'])),
            vectors['leaf_hashes']['B'][1]
        )
        self.assertEqual(
            db32enc(self.proto.hash_leaf(0, leaves['C'])),
            vectors['leaf_hashes']['C'][0]
        )
        self.assertEqual(
            db32enc(self.proto.hash_leaf(1, leaves['C'])),
            vectors['leaf_hashes']['C'][1]
        )

    def test_hash_root(self):
        self.check_hash_root(self.proto)
        self.assertEqual(self._checks, [
            'hash_root_validation',
            'hash_root_simple',
            'hash_root_crypto',
        ])

        vectors = misc.load_data('V1')
        A0 = db32dec(vectors['leaf_hashes']['A'][0])
        A1 = db32dec(vectors['leaf_hashes']['A'][1])
        B0 = db32dec(vectors['leaf_hashes']['B'][0])
        B1 = db32dec(vectors['leaf_hashes']['B'][1])
        C0 = db32dec(vectors['leaf_hashes']['C'][0])
        C1 = db32dec(vectors['leaf_hashes']['C'][1])
        self.assertEqual(
            db32enc(self.proto.hash_root(1, A0)),
            vectors['root_hashes']['A']
        )
        self.assertEqual(
            db32enc(self.proto.hash_root(EightMiB - 1, B0)),
            vectors['root_hashes']['B']
        )
        self.assertEqual(
            db32enc(self.proto.hash_root(EightMiB , C0)),
            vectors['root_hashes']['C']
        )
        self.assertEqual(
            db32enc(self.proto.hash_root(EightMiB + 1, C0 + A1)),
            vectors['root_hashes']['CA']
        )
        self.assertEqual(
            db32enc(self.proto.hash_root(2 * EightMiB - 1, C0 + B1)),
            vectors['root_hashes']['CB']
        )
        self.assertEqual(
            db32enc(self.proto.hash_root(2 * EightMiB , C0 + C1)),
            vectors['root_hashes']['CC']
        )

    def test_vectors(self):
        V1 = misc.load_data('V1')
        self.assertEqual(misc.build_vectors(self.proto, db32enc), V1)

    def test_hex_vectors(self):
        V1_hex = misc.load_data('V1_hex')
        self.assertEqual(misc.build_vectors(self.proto, misc.tohex), V1_hex)

    def test_md5sums(self):
        MD5SUMS = misc.load_data('MD5SUMS')
        self.assertEqual(misc.build_md5sums(self.proto.leaf_size), MD5SUMS)

